/* lightMutex.h
 */
#ifndef OSL_LIGHT_MUTEX_H
#define OSL_LIGHT_MUTEX_H

#include "osl/oslConfig.h"
#include <boost/thread.hpp>
#include <boost/utility.hpp>

namespace osl
{
  namespace misc
  {
#ifdef OSL_USE_RACE_DETECTOR
    typedef boost::mutex LightMutex;
#else
    class LightMutex : boost::noncopyable {
      volatile int lock;
    public:
      class scoped_lock;
      class unlockable_lock;
      LightMutex() :lock(0) {}
      bool tryLock(){
	int dummy;
	asm __volatile__(" movl $1,%0" "\n\t"
			 " xchgl (%1),%0" "\n\t"
			 : "=&q"(dummy)
			 : "q"(&lock)
			 : "cc");
	return dummy==0;
      }
      bool waitLock(int counter){
	for(int i=0;i<counter;i++){
	  asm __volatile__(" pause" "\n\t");
	  if(lock==0)
	    return true;
	}
	return false;
      }
      void unLock(){
	lock=0;
      }
    };
    class LightMutex::scoped_lock : boost::noncopyable {
      LightMutex& m;
    public:
      scoped_lock(LightMutex& m) :m(m){
	while(!m.tryLock()){
	  for(int i=0;i<2;i++){
	    if(!m.waitLock(100)) break;
	    if(m.tryLock()) return;
	  }
	  boost::thread::yield();
	}
      }
      ~scoped_lock(){
	m.unLock();
      }
    };
    /** requirement: thread local */
    class LightMutex::unlockable_lock : boost::noncopyable {
      LightMutex& m;
      bool locked;
    public:
      unlockable_lock(LightMutex& m) :m(m), locked(true) {
	while(!m.tryLock()){
	  for(int i=0;i<2;i++){
	    if(!m.waitLock(100)) break;
	    if(m.tryLock()) return;
	  }
	  boost::thread::yield();
	}
      }
      ~unlockable_lock(){
	unlock();
      }
      void unlock()
      {
	if (locked) {
	  locked = false;
	  m.unLock();
	}
      }
    };
#endif
  }
  using misc::LightMutex;
}
#endif /* OSL_LIGHT_MUTEX_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:

