/* alphaBeta2.cc
 */
#include "osl/search/alphaBeta2.h"
#ifdef OSL_SMP
#  include "osl/search/alphaBeta2Parallel.h"
#endif
#include "osl/search/simpleHashRecord.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/dominanceCheck.h"
#include "osl/search/moveGenerator.h"
#include "osl/search/realizationProbability.h"
#include "osl/search/quiescenceSearch2.h"
#include "osl/search/realizationProbability.h"
#include "osl/search/usiReporter.h"
#include "osl/search/moveWithComment.h"
#include "osl/eval/see.h"
#include "osl/eval/pieceEval.h"
#include "osl/checkmate/immediateCheckmate.h"
#include "osl/record/csa.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/check_.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/move_generator/legalMoves.h"
#include "osl/effect_util/additionalEffect.h"
#include "osl/misc/nonBlockDelete.h"
#include "osl/misc/ctime.h"
#include "osl/stat/ratio.h"
#include "osl/enter_king/enterKing.h"
#include "osl/search/moveStackRejections.h"
#include <stdexcept>
#include <iostream>
#include <iomanip>

#define search_assert(x, m) assert((x) || SearchState2::abort(m))

typedef osl::search::RealizationProbability Probabilities_t;

#ifdef CHECKMATE_COUNT
static size_t root_checkmate = 0, checkmate_before = 0, checkmate_after = 0,
  count_threatmate = 0, quiesce_checkmate=0;
#endif

// #define EXPERIMENTAL_QUIESCE


/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <class EvalT>
osl::CArray<int, osl::search::SearchState2Core::MaxDepth> osl::search::AlphaBeta2Tree<EvalT>::depth_node_count;

template <class EvalT>
osl::search::AlphaBeta2Tree<EvalT>::
AlphaBeta2Tree(const NumEffectState& s, checkmate_t& c, 
	       SimpleHashTable *t, CountRecorder& r)
  : SearchBase<EvalT,SimpleHashTable,CountRecorder,RealizationProbability>(r, t), 
    SearchState2(s, c), AlphaBeta2Common<EvalT>(s), node_count(0), shared_root(new AlphaBeta2SharedRoot)
{
#ifdef OSL_SMP
  for (int i=0; i<4; ++i) {
    try 
    {
      shared.reset(new AlphaBeta2Parallel<EvalT>(this));
      break;
    }
    catch (std::bad_alloc&)
    {
      std::cerr << "panic " << i << " allocation of AlphaBeta2Parallel failed\n";
#ifdef _WIN32
      boost::this_thread::sleep(boost::posix_time::seconds(1));
#else
      sleep(1);
#endif
      NonBlockDelete::deleteAll();
    }
  }
#endif
}

template <class EvalT>
osl::search::AlphaBeta2Tree<EvalT>::
AlphaBeta2Tree(const AlphaBeta2Tree<EvalT>& src, AlphaBeta2Parallel<EvalT> *)
  : SearchBase<EvalT,SimpleHashTable,CountRecorder,RealizationProbability>(src), 
    SearchState2(src), SearchTimer(src), AlphaBeta2Common<EvalT>(src), 
    node_count(0), shared(src.shared), shared_root(src.shared_root)
{
  for (size_t i=0; i<pv.size(); ++i)
    pv[i].clear();
}

template <class EvalT>
osl::search::AlphaBeta2Tree<EvalT>::
~AlphaBeta2Tree()
{
  for (size_t i=0; i<generators.size(); ++i)
    dealloc(generators[i]);
#ifdef OSL_SMP
  if (shared && shared.use_count() == 1)
    NonBlockDelete::reset(shared);
#endif
}

template <class EvalT>
osl::search::MoveGenerator *
osl::search::AlphaBeta2Tree<EvalT>::alloc()
{
  try 
  {
    return new MoveGenerator;
  }
  catch (std::bad_alloc&)
  {
    std::cerr << "panic. allocation of MoveGenerator failed\n";
    throw TableFull();		// stop search anyway
  }
  return 0;
}

template <class EvalT>
void osl::search::AlphaBeta2Tree<EvalT>::dealloc(MoveGenerator *p)
{
  delete p;
}

template <class EvalT>
osl::search::MoveGenerator& osl::search::AlphaBeta2Tree<EvalT>::makeGenerator()
{
  const size_t cur_depth = this->curDepth();
  while (generators.size() <= cur_depth)
    generators.push_back(0);
  if (generators[cur_depth] == 0)
    generators[cur_depth] = alloc();
  return *generators[cur_depth];
}

/* ------------------------------------------------------------------------- */
/*	Methods for alpha beta search					     */
/* ------------------------------------------------------------------------- */

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
alphaBetaSearchAfterMove(const MoveLogProb& moved, Window w,
			 bool in_pv)
{
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
  assert(alt(P) == state().getTurn());
  assert(P == moved.getMove().player());
  assert(eval::notLessThan(P, w.beta(P), w.alpha(P)));

  // PؤPβ => ľμ꤬ˡ
  if (state().inCheck(P)) {
    return this->minusInfty(P);
  }
  this->eval.update(state(), moved.getMove());
  const Player Turn = PlayerTraits<P>::opponent;
  const size_t previous_node_count = nodeCount();

  pv[this->curDepth()].clear();

  int result;
  // ɽԲǽξˤδؿȻ¹Ѥrecord
  // TODO: class  max_depth ٤
  boost::scoped_ptr<SimpleHashRecord> record_if_unavailable;
  SimpleHashRecord *record 
    = this->table->allocate(currentHash(), curLimit());
  const bool has_table_record = record;
  if (! record) {
    record_if_unavailable.reset(new SimpleHashRecord());
    record = record_if_unavailable.get(); // memorize ƤϤʤ
  }
  setCurrentRecord(record);
  record->setInCheck(state().inCheck());
#if 0
  // harmful now
  if (pass_count.loopByBothPass()) {
    return quiesce<Turn>(w);
  }
#endif
  ++node_count;
  int consumption = moved.getLogProb();
  
  // hash μ꤬ʤiid
  if (in_pv 
      && (! record->bestMove().getMove().isNormal())) {
    assert(node_type[this->curDepth()] == PvNode);
    for (int limit = curLimit() - 200+1; limit > consumption+200;
	 limit -= 200) {
      searchAllMoves<Turn>(moved.getMove(), limit, 
			   record, w);
      if (! record->bestMove().validMove()) {
	Move quiesce_best = record->qrecord.bestMove();
	if (quiesce_best.isNormal())
	  record->setBestMove(quiesce_best);
      }
    }
  }
  if (! in_pv) {
    // null window search
    if (node_type[this->curDepth()] == PvNode)
      node_type[this->curDepth()] = AllNode;
    result = searchAllMoves<Turn>
      (moved.getMove(), consumption, record, 
       Window(w.alpha(P), w.alpha(P)));
  } else {
    // normal search
    assert(node_type[this->curDepth()] == PvNode);
    result = searchAllMoves<Turn>(moved.getMove(), consumption, 
				  record, w);
  }
  bool extended = false;
  // extension if the alpha value is updated
  if (eval::betterThan(P, result, w.alpha(P))) {
    const SimpleHashRecord *parent = lastRecord(1);
    int consumption_here = consumption+1;
    if (! w.null() && (! in_pv || consumption > 100))
      consumption_here = std::min(consumption, 100);
    else if (consumption > 100
	     && (record->threatmate().status(P).status() == ThreatmateState::CHECK_AFTER_THREATMATE
		 || record->threatmate().mayHaveCheckmate(P)))
      consumption_here = 100;
    else if (consumption > 150
	     && ((parent && parent->inCheck())
		 || state().hasEffectBy(P, state().getKingPosition(alt(P)))))
      consumption_here = 150;
    if (consumption_here <= consumption) {
      node_type[this->curDepth()] = PvNode;
      extended = true;
      ext_limit.add(consumption - consumption_here);
      result = searchAllMoves<Turn>(moved.getMove(), consumption_here, record, w);
    }
  }
  ext.add(extended);

  if (has_table_record)
    record->addNodeCount(nodeCount() - previous_node_count);
  return result;
}

template <class EvalT>
template <osl::Player Turn>
int osl::search::AlphaBeta2Tree<EvalT>::
searchAllMoves(Move m, int limit_consumption, SimpleHashRecord *record, 
	       Window w)
{
  if (! w.null())
    assert(node_type[this->curDepth()] == PvNode);
  const Player P = PlayerTraits<Turn>::opponent;
  this->recorder.tryMove(MoveLogProb(m, limit_consumption),
		   w.alpha(P), curLimit());
  subLimit(limit_consumption);

  const int result = searchAllMoves<Turn>(record, w);

  addLimit(limit_consumption);
  this->recorder.recordValue(MoveLogProb(m, limit_consumption),
			     result,eval::betterThan(P, result, w.alpha(P)),
		       curLimit());
  return result;
}

template <class EvalT>
template <osl::Player P>
void osl::search::AlphaBeta2Tree<EvalT>::
testThreatmate(SimpleHashRecord *record, bool in_pv)
{
  if ((! record->inCheck())
      && (! (record && record->threatmate().isThreatmate(P)))
      && tryThreatmate())
  {
    int threatmate_limit = 0;

    threatmate_limit = 4500-this->curDepth()*1000;
    threatmate_limit = std::max(threatmate_limit, 100);
    if (! in_pv)
      threatmate_limit /= 2;
    if (root_limit < 800)
      threatmate_limit /= 2;
#ifdef EXPERIMENTAL_QUIESCE
    if (curLimit() <= 400)
      threatmate_limit = 1;
    else if (curLimit() <= 500)
      threatmate_limit /= 16;
    else if (curLimit() <= 600)
      threatmate_limit /= 4;
#endif

    if (curLimit() >= this->table->minimumRecordLimit())
    {
      threatmate_limit = record->qrecord.threatmateNodesLeft(threatmate_limit);
    }
    else
    {
      threatmate_limit /= 2;		// for safety against multiple calls
    }

    Move threatmate_move;
    this->recorder.gotoCheckmateSearch(state(), threatmate_limit);
#ifdef CHECKMATE_COUNT
    size_t count = checkmateSearcher().totalNodeCount();
#endif
    bool threatmate
      = isThreatmateState<P>(threatmate_limit, threatmate_move, 
			     record->qrecord.threatmate_oracle);
#ifdef CHECKMATE_COUNT
    count_threatmate += checkmateSearcher().totalNodeCount() - count;
#endif
    if (threatmate_limit > 100) {
      updateCheckmateCount();
      testStop();
    }
    this->recorder.backFromCheckmateSearch();
    if (!threatmate && threatmate_limit == 0
	&& record->qrecord.threatmateNodesLeft(2)) {
      threatmate = isThreatmateStateShort<P>(2, threatmate_move);
    }
    if (threatmate)
    {
      record->threatmate().setThreatmate(P, threatmate_move);
    }
  }
}

template <class EvalT>
template <osl::Player P>
bool osl::search::AlphaBeta2Tree<EvalT>::
tryCheckmate(SimpleHashRecord *record, bool in_pv, Move& checkmate_move)
{
  int checkmate_limit = 1;
  if (! in_pv) {
    const SimpleHashRecord *parent = lastRecord(1);
    if (! (record->threatmate().mayHaveCheckmate(alt(P))
	   || (parent && parent->threatmate().maybeThreatmate(alt(P)))))
      return false;
    // simulation only
  }
  if (in_pv && root_limit >= 500) {
    int depth = this->curDepth();
    if (root_limit >= 700 && depth <= 3) {
      if (/*record_memorize &&*/ (depth <= 1))
	checkmate_limit = checkmate::limitToCheckCount(3500);
      else if (/* record_memorize &&*/ (depth == 2))
	checkmate_limit = 1000;
      else if (/* record_memorize &&*/ (depth == 3))
	checkmate_limit = 200;
    }
    else if (((root_limit - curLimit()) <= 500) || (depth <= 5))
    {
      assert(static_cast<unsigned int>(curLimit()) < 4096);
      checkmate_limit = checkmate::limitToCheckCount(std::max(0,curLimit()-200));
    }
    const SimpleHashRecord *parent = lastRecord(1);
    if (record->threatmate().mayHaveCheckmate(alt(P))
	|| (parent && parent->threatmate().maybeThreatmate(alt(P))))
      checkmate_limit += std::max(100, checkmate_limit);
    if (root_limit < 800)
      checkmate_limit /= 2;
  }
  if (curLimit() >= this->table->minimumRecordLimit())
  {
    // ;ϤٿȤƤ
    checkmate_limit = record->qrecord.checkmateNodesLeft(checkmate_limit);
    if (checkmate_limit <= 0)
      return false;
  }
  else
  {
    checkmate_limit /= 2;		// for safety against multiple calls
  }
  
  // ͤޤȤͤ 
#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  this->recorder.gotoCheckmateSearch(state(), checkmate_limit);
  const bool win = isWinningState<P>
    (checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
  if (checkmate_limit > 100)
    updateCheckmateCount();
  this->recorder.backFromCheckmateSearch();
#ifdef CHECKMATE_COUNT
  checkmate_before += checkmateSearcher().totalNodeCount() - count;
#endif
  return win;
}

template <class EvalT>
template <osl::Player P>
bool osl::search::AlphaBeta2Tree<EvalT>::
tryCheckmateAgain(SimpleHashRecord *record, Move& checkmate_move,
		  int node_count, int best_value)
{
  int checkmate_limit = 1;
  if (record->threatmate().maybeThreatmate(P)) {
    if (EvalTraits<P>::betterThan(this->eval.captureValue(newPtypeO(P,KING)), best_value))
      checkmate_limit = node_count / 2; // ζ̤ɬ or ;ʳalt(P)ξ
    else 
      checkmate_limit = node_count / 8;
    checkmate_limit += 100;
  } else if (record->threatmate().mayHaveCheckmate(alt(P))) {
    checkmate_limit = countCheckAfterThreatmate(alt(P),2)*320 + 100;
  }
  if (curDepth() == 1 && hasLastRecord(1)) {
    const SimpleHashRecord *parent = lastRecord(1); // root
    if (parent->inCheck() || parent->threatmate().maybeThreatmate(alt(P)))
      checkmate_limit = std::max(checkmate_limit, (int)parent->nodeCount()/2+parent->checkmateNodes());
  }
  if (true /*record_memorize*/)
    checkmate_limit = record->qrecord.checkmateNodesLeft(checkmate_limit);

#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  this->recorder.gotoCheckmateSearch(state(), checkmate_limit);
  const bool win = isWinningState<P>
    (checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
  if (checkmate_limit > 100)
    updateCheckmateCount();
  this->recorder.backFromCheckmateSearch();
#ifdef CHECKMATE_COUNT
  checkmate_after += checkmateSearcher().totalNodeCount() - count;
#endif

  return win;
}

template <class EvalT>
bool osl::search::
AlphaBeta2Tree<EvalT>::tryPass(SimpleHashRecord *record, Player P) const
{
  return ! record->inCheck()
    && ! record->threatmate().maybeThreatmate(P);
}

template <class EvalT>
template <osl::Player P>
const osl::MoveLogProb osl::search::
AlphaBeta2Tree<EvalT>::nextMove()
{
  MoveGenerator& generator = makeGenerator();
  SimpleHashRecord *record = lastRecord();
  switch (this->move_type[this->curDepth()]) {
  case common_t::HASH:
  {
    if (curLimit() < this->leaf_limit) {
      this->move_type[this->curDepth()] = common_t::FINISH;
      break;
    }
    this->move_type[this->curDepth()] = common_t::TACTICAL;
    const MoveLogProb best_move_in_table = record->bestMove().moveLogProb();
    assert(curLimit() > 0);
    generator.init<eval_t>(curLimit(), record, this->eval, state(), 
			   node_type[this->curDepth()] != CutNode,
			   best_move_in_table.getMove());
    if (best_move_in_table.validMove() && 
	this->validTableMove(state(), best_move_in_table, curLimit())) {
      return best_move_in_table;
    }
  }
  // fall through
  // TODO: ͤϤǥå᤽
  case common_t::TACTICAL:
  {
    MoveLogProb m = generator.nextTacticalMove<P>(*this);
    if (m.validMove())
      return m;
    // fall through
    this->move_type[this->curDepth()] = common_t::KILLER;
    this->killers[this->curDepth()].clear();
    if ((! record->inCheck())
	&& ! record->threatmate().maybeThreatmate(P)
	&& (curLimit() >= 300)) {
      MoveVector killer_moves;	// TODO: Ψ
      getKillerMoves(killer_moves);
      for (MoveVector::const_iterator p=killer_moves.begin(); 
	   p!=killer_moves.end(); ++p) {
	assert(this->killers[this->curDepth()].size() < this->killers[this->curDepth()].capacity());
	this->killers[this->curDepth()].push_back(*p);
      }
      std::reverse(this->killers[this->curDepth()].begin(), this->killers[this->curDepth()].end());
    }
  }
  case common_t::KILLER:
  {
    typename common_t::killer_t& killers = this->killers[this->curDepth()];
    if (! killers.empty()) {
      Move m = killers[killers.size()-1];
      killers.pop_back();
      return MoveLogProb(m, 300);
    }
    // fall through
    this->move_type[this->curDepth()] = common_t::PASS;
  }
  case common_t::PASS:
    assert(record->inCheck() == state().inCheck());
    this->move_type[this->curDepth()] = common_t::ALL;
    if (tryPass(record, P)) {
      const int pass_consumption = (curLimit() >= 800) ? 300 : 200;
      return MoveLogProb(Move::PASS(P), pass_consumption);
    }
    // TODO: pass θκ
    // fall through
  case common_t::ALL:
  {
    MoveLogProb m = generator.nextMove<P>(*this);
    if (m.validMove())
      return m;
    this->move_type[this->curDepth()] = common_t::FINISH;
  }
  default:
    assert(this->move_type[this->curDepth()] == common_t::FINISH);
  }
  return MoveLogProb();
}
  
template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
searchAllMoves(SimpleHashRecord *record, Window w)
{
#ifndef NDEBUG
  checkPointSearchAllMoves();
#endif
  assert(P == state().getTurn());
  search_assert(w.isConsistent(), lastMove());
  assert(curLimit() >= 0);

  assert(hasLastRecord(1));
  const SimpleHashRecord *parent = lastRecord(1);
  const int node_count_at_beginning = nodeCount();
#ifndef OSL_USE_RACE_DETECTOR
  depth_node_count[this->curDepth()]++;
#endif
  this->move_type[this->curDepth()] = common_t::INITIAL;
  const bool in_pv = this->in_pv[this->curDepth()] = ! w.null();

  // ơ֥ˤͤĴ٤
  if (record) {
    if (in_pv) {
      if (record->hasLowerBound(SearchTable::HistorySpecialDepth)) {
	int lower_bound = record->lowerBound();
	if (this->isWinValue(P, lower_bound))
	  return lower_bound;
      }
      if (record->hasUpperBound(SearchTable::HistorySpecialDepth)) {
	int upper_bound = record->upperBound();
	if (this->isWinValue(alt(P), upper_bound))
	  return upper_bound;
      }
    }
    else {			// ! in_pv
      int table_value = 0;
      if (record->hasGreaterLowerBound<P>(curLimit(), w.alpha(P), 
					  table_value)) {
	assert(eval::isConsistentValue(table_value));
	w.alpha(P) = table_value + EvalTraits<P>::delta;
	if (EvalTraits<P>::betterThan(table_value, w.beta(P))) {
	  this->recorder.tableHitLowerBound(P, table_value, w.beta(P), curLimit());
	  return table_value;
	}
      } 
      if (record->hasLesserUpperBound<P>(curLimit(), w.beta(P), table_value)) {
	assert(eval::isConsistentValue(table_value));
	w.beta(P) = table_value - EvalTraits<P>::delta;
	if (EvalTraits<P>::betterThan(w.alpha(P), table_value))
	{
	  this->recorder.tableHitUpperBound(P, table_value, w.alpha(P), curLimit());
	  return table_value;
	}
      }
    }

    Move checkmate_move=Move::INVALID();
    if ((!record->inCheck())
	&& record->qrecord.checkmateNodesLeft(1)
	&& isWinningStateShort<P>(2, checkmate_move))
    {
      this->recordWinByCheckmate(P, record, checkmate_move);
      return this->winByCheckmate(P);
    }    
#ifndef DONT_USE_CHECKMATE
    assert(record);
    // try simulation or simple checkmate search
    int additional_limit = 0;		// simulation only
    if (parent && parent->threatmate().maybeThreatmate(alt(P)))
    {
      additional_limit = std::max(100, parent->qrecord.threatmateNodes()/8);
      additional_limit = record->qrecord.checkmateNodesLeft(additional_limit);
    }
    this->recorder.gotoCheckmateSearch(state(), additional_limit);
    const bool win = isWinningState<P>(additional_limit, checkmate_move, 
				       record->qrecord.attack_oracle);
    updateCheckmateCount();
    this->recorder.backFromCheckmateSearch();
    if (win) {      
      assert(checkmate_move.isValid());
      this->recordWinByCheckmate(P, record, checkmate_move);
      return this->winByCheckmate(P);
    }
#endif
  }

  search_assert(w.isConsistent(), lastMove());
  const int initial_alpha = w.alpha(P);

#ifndef DONT_USE_CHECKMATE
  // ͤͤ
  testThreatmate<P>(record, in_pv);
#endif
  // õ ThreatmateState 
  record->threatmate().update(P, (parent ? &(parent->threatmate()) : 0), 
			      state().inCheck());

  MoveLogProb best_move;	// invalidated
  int best_value = this->minusInfty(P);
  int tried_moves = 0;
  int alpha_update = 0;
  int last_alpha_update = 0;
#if (defined OSL_SMP) 
  int last_smp_idle = 0;
#  if (! defined OSL_SMP_NO_SPLIT_INTERNAL)
  bool already_split = false;
#  endif
#endif

  // the first move
  MoveLogProb m = nextMove<P>();
  ++tried_moves;
  if (! m.validMove() || m.getLogProb() > curLimit()) {
    goto move_generation_failure;
  }
  int first_move_node;
  {
    const int previous_node_count = nodeCount();

    assert(eval::betterThan(P, w.alpha(P), best_value));
    const int result = alphaBetaSearch<P>(m, w, in_pv);
    if (eval::betterThan(P, result, best_value)) {
      best_value = result;
      best_move = m;
      if (eval::betterThan(P, best_value, w.alpha(P))) {
	w.alpha(P) = result + EvalTraits<P>::delta;;
	assert(best_move.validMove());
	++alpha_update;
	last_alpha_update = 1;
	if (eval::betterThan(P, result, w.beta(P))) {
	  mpn_cut.add(tried_moves);
	  if (this->move_type[this->curDepth()] >= common_t::ALL)
	    setKillerMove(best_move.getMove());
	  assert(best_move.validMove());
	  assert(! this->isWinValue(alt(P), best_value));
	  goto register_table;
	} else {
	  if (in_pv) 
	    makePV(m.getMove());
	}
      }
    }
    first_move_node = nodeCount() - previous_node_count;
  }
  testStop();

#ifndef DONT_USE_CHECKMATE
  if (in_pv)
  {
    Move checkmate_move;
    if (tryCheckmate<P>(record, in_pv, checkmate_move)) {
      assert(checkmate_move.isValid());
      best_value= this->winByCheckmate(P);
      this->recordWinByCheckmate(P, record, checkmate_move);
      return best_value;
    }
  }
#endif
  search_assert(w.isConsistent(), lastMove());
  if (curLimit() < this->leaf_limit)
    goto move_generation_failure;
  while (true) {
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_INTERNAL)
    const bool prefer_split = shared && curLimit() >= shared->split_min_limit
      && (curLimit() >= 600
	  // || (curLimit() >= 400 && this->curDepth() < 2)
	  || (first_move_node >= 30000));
    try {
      if (prefer_split) {
	int cur_smp_idle; 
	{
# ifdef OSL_USE_RACE_DETECTOR
	  boost::mutex::scoped_lock lk(shared->lock_smp);
#endif
	  cur_smp_idle = shared->smp_idle;
	}
	if (cur_smp_idle > last_smp_idle) {
	  last_smp_idle = cur_smp_idle;
	  assert(! already_split);
	  already_split = true;
	  if (examineMovesOther<P>(w, best_move, best_value, tried_moves,
				   alpha_update, last_alpha_update)) {
	    assert(best_move.validMove());
	    assert(best_move.getMove().player() == P);
	    if (this->move_type[this->curDepth()] >= common_t::ALL)
	      setKillerMove(best_move.getMove());
	    mpn_cut.add(tried_moves);    
	    goto register_table;
	  }
	  goto all_moves_done;
	}
      }
    }
    catch(AlphaBeta2ParallelCommon::SplitFailed&) {
      already_split = false;
      // fall through
    }
#endif
    MoveLogProb m = nextMove<P>(); 
    if (! m.validMove())
      break;
    ++tried_moves;

    assert(eval::betterThan(P, w.alpha(P), best_value));
    const int result = alphaBetaSearch<P>(m, w, in_pv && ! best_move.validMove());
    if (eval::betterThan(P, result, best_value)) {
      best_value = result;
      best_move = m;
      if (eval::betterThan(P, best_value, w.alpha(P))) {
	w.alpha(P) = result + EvalTraits<P>::delta;;
	assert(best_move.validMove());
	++alpha_update;
	last_alpha_update = tried_moves;
	if (eval::betterThan(P, result, w.beta(P))) {
	  assert(best_move.validMove());
	  if (this->move_type[this->curDepth()] >= common_t::ALL)
	    setKillerMove(best_move.getMove());
	  mpn_cut.add(tried_moves);
	  goto register_table;
	} else {
	  if (in_pv)
	    makePV(m.getMove());
	}
      }
    }
  }
#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_INTERNAL)
all_moves_done:
#endif
  if (tried_moves == 1 && tryPass(record, P)) {
    // goto quiescence search if tried move is only null move
    goto move_generation_failure;
  }
  mpn.add(tried_moves);
  if (alpha_update) {
    this->alpha_update.add(alpha_update);
    this->last_alpha_update.add(last_alpha_update);
  }
  // 
  if (((this->curDepth() % 2) == 0)
      && EnterKing::canDeclareWin<P>(state())) {
    best_value = this->brinkmatePenalty(alt(P), std::max(1,16-this->curDepth())*256)
      + this->eval.value();
    record->setAbsoluteValue(Move::DeclareWin(), best_value,
			     SearchTable::CheckmateSpecialDepth);
    return best_value;
  }
  if (record) {
    // ⤦١ͤޤȤͤ
#ifndef DONT_USE_CHECKMATE
    Move checkmate_move=Move::INVALID();
    if (tryCheckmateAgain<P>(record, checkmate_move, 
			     nodeCount() - node_count_at_beginning,
			     best_value)) {
      assert(checkmate_move.isValid());
      best_value= this->winByCheckmate(P);
      this->recordWinByCheckmate(P, record, checkmate_move);
      return best_value;
    }
#endif
  }
register_table:
  assert(best_value == this->minusInfty(P) || best_move.validMove());
  assert(eval::isConsistentValue(best_value));
  if (this->isWinValue(alt(P), best_value))
  {
    // TODO: ľ꤬(䤱)Ϣ³ä
    // ɬǤϤʤƵͰɤǽϤ
    best_value = this->brinkmatePenalty(P, std::max(1,16-this->curDepth())*256) + this->eval.value();

    // (οǤ)==
    record->setAbsoluteValue(best_move, best_value, curLimit());
    return best_value;
  }
  else if (EvalTraits<P>::betterThan(w.alpha(P), initial_alpha)) {
    if (best_move.validMove()) {
      best_move.setLogProb(RealizationProbability::TableMove);
      assert(best_value % 2 == 0);
      record->setLowerBound(P, curLimit(), 
			    SearchMove(best_move), best_value);
    }
  }
  if (EvalTraits<P>::betterThan(w.beta(P), best_value)) {
    if (best_move.validMove()) {
      best_move.setLogProb(RealizationProbability::TableMove);
      record->setUpperBound(P, curLimit(), 
			    SearchMove(best_move), best_value);
    }
  }
  return best_value;
move_generation_failure:
  pv[this->curDepth()].clear();
  // Ǥʤä
  // limit  200 ̤äȡʾä move generator  limit ʲμǤʤä礳
  best_value = quiesce<P>(w);
  if (record)
  {
    if (EvalTraits<P>::betterThan(best_value, initial_alpha)) {
      if (EvalTraits<P>::betterThan(w.beta(P), best_value)) {
	record->setAbsoluteValue(MoveLogProb(), best_value, curLimit());
      } else {
	this->recordLowerBound(P, record, curLimit(), SearchMove(), best_value);
      }
    }
    else 
    {
      assert(EvalTraits<P>::betterThan(w.beta(P), best_value));
      this->recordUpperBound(P, record, curLimit(), SearchMove(), best_value);
    }
  }
  assert(eval::isConsistentValue(best_value));
  return best_value;
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
quiesce(Window w)
{
#ifdef EXPERIMENTAL_QUIESCE
  return quiesceExp<P>(w);
#else
  return quiesceStable<P>(w);
#endif
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
quiesceStable(Window w)
{
  testStop();
  initPV();
  
  typedef QuiescenceSearch2<eval_t> qsearcher_t;
  qsearcher_t qs(*this, *this->table);
  Move last_move = lastMove();
  if (last_move.isInvalid())
    last_move = Move::PASS(alt(P));
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
#ifdef CHECKMATE_COUNT
  size_t count = checkmateSearcher().totalNodeCount();
#endif
  const int result = qs.template search<P>(w.alpha(P), w.beta(P), this->eval, last_move, 4);
  node_count += qs.nodeCount();
  this->recorder.addQuiescenceCount(qs.nodeCount());
#ifdef CHECKMATE_COUNT
  quiesce_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif

  assert(result % 2 == 0);
  return result;
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
quiesceExp(Window w)
{
  testStop();

  SimpleHashRecord *record = lastRecord();
  assert(record);
  Move best_move;
  const int qdepth = 4;
  const int previous_node_count = nodeCount();

  int result =
    quiesceRoot<P>(w, qdepth, best_move, record->threatmate());
  
  const size_t qnode = nodeCount() - previous_node_count;
  this->recorder.addQuiescenceCount(qnode);
  record->qrecord.setLowerBound(qdepth, result, best_move);
  return result;
}

template <class EvalT>
template <osl::Player P>
struct osl::search::AlphaBeta2Tree<EvalT>::NextQMove
{
  AlphaBeta2Tree *searcher;
  Window window;
  const int depth;
  int *result;
  DualThreatmateState threatmate;
  NextQMove(AlphaBeta2Tree *s, Window w, int d, int *r,
	    DualThreatmateState t)
    : searcher(s), window(w), depth(d), result(r), threatmate(t) {
  }
  void operator()(Position /*last_to*/) {
    searcher->eval.update(searcher->state(), searcher->lastMove());
    *result = 
      searcher->quiesce<P>(window, depth, threatmate);
  }
};

template <class EvalT>
template <osl::Player P>
bool osl::search::AlphaBeta2Tree<EvalT>::
quiesceWithMove(Move move, Window& w, int depth_left, Move& best_move, int& best_value, 
		const DualThreatmateState& threatmate)
{
  // TODO: futility margin
  const bool in_pv = ! w.null();
  int result;
  typedef NextQMove<PlayerTraits<P>::opponent> next_t;
  next_t helper(this, w, depth_left, &result, threatmate);

  const HashKey new_hash = currentHash().newHashWithMove(move);
  const eval_t old_eval = this->eval;
  doUndoMoveOrPass<P,next_t>(new_hash, move, helper);
  this->eval = old_eval;

  if (eval::betterThan(P, result, best_value)) {
    best_value = result;
    best_move = move;
    if (eval::betterThan(P, best_value, w.alpha(P))) {
      w.alpha(P) = result + EvalTraits<P>::delta;
      if (in_pv)
	makePV(best_move);
      if (eval::betterThan(P, result, w.beta(P))) {
	return true;
      }
    }
  }
  return false;
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
quiesceRoot(Window w, int depth_left, Move& best_move, DualThreatmateState threatmate)
{
  assert(! state().inCheck(alt(P)));

  initPV();
  // depth_node_count_quiesce[this->curDepth()]++;
  // ++node_count;

  SimpleHashRecord& record = *lastRecord();
  assert(record.inCheck() == state().inCheck());
  assert(depth_left > 0);
  
  int best_value = this->minusInfty(P);
  // stand pat 
  if (! record.inCheck()) {
    if (! threatmate.maybeThreatmate(P)) {
      best_value = this->eval.value();
    } else {
      const int value = this->eval.value() + this->threatmatePenalty(P);
      best_value = EvalTraits<P>::max(best_value, value);
    }
    best_move = Move::PASS(P);
    if (EvalTraits<P>::betterThan(best_value, w.alpha(P))) {
      if (EvalTraits<P>::betterThan(best_value, w.beta(P)))
	return best_value;
      w.alpha(P) = best_value + EvalTraits<P>::delta;
    }
  }

  Move prev_best = record.qrecord.bestMove();
  MoveGenerator& generator = makeGenerator();
  generator.init(200, &record, this->eval, state(), 
		 w.alpha(P) == w.beta(P),
		 prev_best, true);
  int tried_moves = 0;
  // previous 
  if (prev_best.isNormal()) {
    ++tried_moves;
    if (quiesceWithMove<P>(prev_best, w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  }


  // captures or king escape
  for (MoveLogProb m = generator.nextTacticalMove<P>(*this); 
       m.validMove(); m = generator.nextTacticalMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 
  for (MoveLogProb m = generator.nextMove<P>(*this); 
       m.validMove(); m = generator.nextMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 

  // pawn drop foul?
  if (record.inCheck()) {
    if (tried_moves == 0) {
      if (lastMove().isNormal() && lastMove().ptype() == PAWN && lastMove().isDrop())
	return this->winByFoul(P);
      return this->winByCheckmate(alt(P));
    }
    goto finish;
  }  
finish:
  return best_value;
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
quiesce(Window w, int depth_left, DualThreatmateState parent_threatmate)
{
  if (state().inCheck(alt(P))) {
    return this->minusInfty(alt(P));
  }

  initPV();
  depth_node_count_quiesce[this->curDepth()]++;
  ++node_count;

  SimpleHashRecord record;
  record.inCheck() = state().inCheck();

  DualThreatmateState threatmate;
  threatmate.update(P, &parent_threatmate, record.inCheck());

  int best_value = this->minusInfty(P);
  // TODO: ̤βtakebackĹ
  if (depth_left <= 0) {
    if (record.inCheck()) {
      if (lastMove().capturePtype() != PTYPE_EMPTY)
	depth_left +=2;
      else
	depth_left = 0;
    }
    else if (threatmate.maybeThreatmate(P)) {
      if (threatmate.mayHaveCheckmate(alt(P))) {
	Move checkmate_move;
	bool win = isWinningState<P>(10, checkmate_move, record.qrecord.attack_oracle);
	if (win)
	  return this->winByCheckmate(P);
      }
      return this->eval.value() + this->threatmatePenalty(P);
    }
    else {
      if (threatmate.mayHaveCheckmate(alt(P)))
	return this->eval.value() + this->threatmatePenalty(alt(P));
      if (ImmediateCheckmate::hasCheckmateMove<P>(state()))
	return this->winByCheckmate(P);
      if (ImmediateCheckmate::hasCheckmateMove<PlayerTraits<P>::opponent>(state()))
	return this->eval.value() + this->threatmatePenalty(P);
      return this->eval.value();
    }
  }

  if (! record.inCheck()) {
    if (ImmediateCheckmate::hasCheckmateMove<P>(state())) {
      return this->winByCheckmate(P);
    }
  }
  if (threatmate.mayHaveCheckmate(alt(P))) {
    Move checkmate_move;
    bool win = isWinningState<P>(10, checkmate_move, record.qrecord.attack_oracle);
    if (win)
      return this->winByCheckmate(P);
  }
  MoveGenerator& generator = makeGenerator();
  generator.init(200, &record, this->eval, state(), 
		 w.alpha(P) == w.beta(P),
		 Move(), true);
  int tried_moves = 0;
  Move best_move;
  // stand pat
  if (! record.inCheck()) {
    if (! threatmate.maybeThreatmate(P)) {
      best_value = this->eval.value();
    } else {
      const int value = this->eval.value() + this->threatmatePenalty(P);
      best_value = EvalTraits<P>::max(best_value, value);
    }
    best_move = Move::PASS(P);
    if (EvalTraits<P>::betterThan(best_value, w.alpha(P))) {
      if (EvalTraits<P>::betterThan(best_value, w.beta(P)))
	return best_value;
      w.alpha(P) = best_value + EvalTraits<P>::delta;
    }
  }

  // captures or king escape
  for (MoveLogProb m = generator.nextTacticalMove<P>(*this); 
       m.validMove(); m = generator.nextTacticalMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 
  for (MoveLogProb m = generator.nextMove<P>(*this); 
       m.validMove(); m = generator.nextMove<P>(*this)) {
    ++tried_moves;
    if (quiesceWithMove<P>(m.getMove(), w, depth_left-1, best_move, best_value,
			   threatmate))
      goto finish;
  } 

  // pawn drop foul?
  if (record.inCheck()) {
    if (tried_moves == 0) {
      if (lastMove().ptype() == PAWN && lastMove().isDrop()) 
	return this->winByFoul(P);
      return this->winByCheckmate(alt(P));
    }
    goto finish;
  }  
finish:
  return best_value;
}

template <class EvalT>
void osl::search::AlphaBeta2Tree<EvalT>::updateCheckmateCount()
{
#ifdef OSL_SMP
  if (shared) {
    this->recorder.setCheckmateCount(shared->checkmateCount());
    return;
  }
#endif
  this->recorder.setCheckmateCount
    (checkmateSearcher().totalNodeCount());
}

template <class EvalT>
int 
osl::search::AlphaBeta2Tree<EvalT>::
rootAlpha(Player P, int last_value)
{
  int pawns = 3;
  if (eval::betterThan(P, last_value, eval_t::captureValue(newPtypeO(alt(P),KING))))
    pawns = 10;
  const int width = eval_t::captureValue(newPtypeO(alt(P),PAWN))*pawns/2;
  return last_value - width - eval::delta(P);
}

template <class EvalT>
int 
osl::search::AlphaBeta2Tree<EvalT>::
stableThreshold(Player P, int last_value)
{
  int pawns = 3;
  if (eval::betterThan(P, last_value, eval_t::captureValue(newPtypeO(alt(P),KING))))
    pawns = 10;
  else if (eval::betterThan(P, eval_t::captureValue(newPtypeO(alt(P),PAWN))*2, last_value)
	   && eval::betterThan(P, last_value, eval_t::captureValue(newPtypeO(P,PAWN))*2))
    pawns = 2;
  const int width = eval_t::captureValue(newPtypeO(alt(P),PAWN))*pawns/2;
  return last_value - width - eval::delta(P);
}

template <class EvalT>
void osl::search::AlphaBeta2Tree<EvalT>::
updateRootPV(Player P, std::ostream& os, int result, Move m)
{
#ifdef OSL_SMP
  boost::scoped_ptr<boost::mutex::scoped_lock> lk;
  if (shared)
    lk.reset(new boost::mutex::scoped_lock(shared->lock_io));
#endif
  this->makePV(m);
  const int last_root_value = shared_root->last_root_value.size() ? shared_root->last_root_value.back() : 0;
  const int threshold = stableThreshold(P, last_root_value);
  bool new_stable = eval::betterThan(P, result, threshold);
  shared_root->last_root_value_update = result;

  if (new_stable && m != shared_root->last_root_move
      && (See::see(state(), m) < -eval::Ptype_Eval_Table.value(KNIGHT)*2
	  || eval::betterThan(P, result, eval_t::captureValue(newPtypeO(alt(P),KING))))) {
    new_stable = false;
  }
  if (new_stable && shared_root->last_root_value.size() > 1) {
    const int last_root_value2 = shared_root->last_root_value[shared_root->last_root_value.size()-2];
    const int threshold2 = stableThreshold(P, last_root_value2);
    if (eval::betterThan(P, threshold2, result)
	&& eval::betterThan(P, last_root_value2, last_root_value))
      new_stable = false;
  }
  this->shared_root->last_pv = pv[0];
  this->setStable(new_stable);
#ifndef MINIMAL
  const double scale = 200.0/eval_t::captureValue(newPtypeO(alt(P),PAWN));
  if (OslConfig::usiMode() && !this->root_ignore_moves)
    UsiReporter::showPV(root_limit/200, this->nodeCount(), this->elapsed(),
			result*scale, 
			m, &*pv[0].begin(), &*pv[0].end());
#endif
  if (this->table->isVerbose()) {
    showPV(std::cerr, result, m, new_stable);
  }
}

template <class EvalT>
void osl::search::AlphaBeta2Tree<EvalT>::
showFailLow(int result, Move m) const
{
  if (this->root_ignore_moves)
    std::cerr << "[" << this->root_ignore_moves->size() << "] ";
  std::cerr << " <" << std::setfill(' ') << std::setw(5) 
	    << static_cast<int>(result*200.0/this->eval.captureValue(newPtypeO(WHITE,PAWN)))
	    << " " << record::csa::show(m) << "\n";
}

template <class EvalT>
void osl::search::AlphaBeta2Tree<EvalT>::
showPV(std::ostream& os, int result, Move m, bool stable) const
{
  assert(m.isNormal()); // do not pass at root
  if (this->root_ignore_moves)
    os << "[" << this->root_ignore_moves->size() << "] ";
  os << (stable ? ' ' : '*');
  os << " " << std::setfill(' ') << std::setw(5) 
     << static_cast<int>(result*200.0/this->eval.captureValue(newPtypeO(WHITE,PAWN))) << " ";
  for (size_t i=0; i<pv[0].size(); ++i) {
    os << record::csa::show(pv[0][i]);
  }
  os << std::endl;

#ifdef DEBUG_PV
  NumEffectState s = state();
  for (size_t i=0; i<pv[0].size(); ++i) {
    if (! pv[0][i].isPass() && ! s.isValidMove(pv[0][i])) {
      std::cerr << "root pv error " << pv[0][i] << " " << i << "\n";
      break;
    }
    ApplyMoveOfTurn::doMove(s, pv[0][i]);
  }
#endif
}

template <class EvalT>
template <osl::Player P>
class osl::search::AlphaBeta2Tree<EvalT>::NextMove
{
public:
  AlphaBeta2Tree *searcher;
  const MoveLogProb& moved;
  Window window;
  int *result;
  bool in_pv;
  NextMove(AlphaBeta2Tree *s, const MoveLogProb& md, Window w, int *r,
	   bool p)
    : searcher(s), moved(md), window(w), result(r), in_pv(p) {
    assert(P == md.getMove().player());
  }
  void operator()(Position /*last_to*/) {
#ifndef NDEBUG
    const int cur_limit = searcher->curLimit();
#endif
    *result = 
      searcher->alphaBetaSearchAfterMove<P>(moved, window, in_pv);
    assert(cur_limit == searcher->curLimit() || searcher->SearchState2Core::abort());
  }
};

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2Tree<EvalT>::
alphaBetaSearch(const MoveLogProb& search_move, Window w, bool in_pv)
{
  assert(w.alpha(P) % 2);
  assert(w.beta(P) % 2);
  const Move move = search_move.getMove();
  assert(P == move.player());
  assert(P == state().getTurn());
  assert(eval::notLessThan(P, w.beta(P), w.alpha(P)));

  testStop();
  pv[curDepth()+1].clear();
  // TODO: more efficient way to find foul
  if (! move.isPass() ){
    if(MoveStackRejections::probe<P>(state(),history(),curDepth(),move,w.alpha(P),repetitionCounter().checkCount(alt(P)))){
      return this->winByLoop(alt(P));
    }
    if (move_classifier::MoveAdaptor<move_classifier::PawnDropCheckmate<P> >
	::isMember(state(), move))
      return this->winByFoul(alt(P));
  }

  const HashKey new_hash = currentHash().newHashWithMove(move);
  assert(P == move.player());

  if (move.isPass())
    this->pass_count.inc(P);

  // ǧ
  if (! this->pass_count.loopByBothPass()) {
    const Sennichite next_sennichite
      = repetition_counter.isAlmostSennichite(new_hash);
    if (next_sennichite.isDraw())
      return this->drawValue();
    if (next_sennichite.hasWinner())
      return this->winByFoul(next_sennichite.winner());
    assert(next_sennichite.isNormal());
  }
  
  if (! move.isPass()) {
    // ͥ۴طǧ
    const DominanceCheck::Result has_dominance
      = DominanceCheck::detect(repetition_counter.history(), new_hash);
    if (has_dominance == DominanceCheck::LOSE)
      return this->winByLoop(alt(P));
    if (has_dominance == DominanceCheck::WIN)
      return this->winByLoop(P);
    // Ϣ³Τ
    if (move.capturePtype() == PTYPE_EMPTY) {
      const int sacrifice_count = countSacrificeCheck2(this->curDepth());
      if (sacrifice_count == 2) {
	// 3ܤϻؤʤ
	const Position to = move.to();
	int offence = state().countEffect(P, to) + (move.isDrop() ? 1 : 0);
	const int deffense = state().hasEffectBy(alt(P), to); // max1
	if (offence <= deffense)
	  offence += AdditionalEffect::count2(state(), to, P);
	if (offence <= deffense) {
	  return this->winByLoop(alt(P));
	}
      }
    }
  }
  // õ
  int result;
  NextMove<P> helper(this, search_move, w, &result, in_pv);

  this->recorder.addNodeCount();  
  const eval_t old_eval = this->eval;
  doUndoMoveOrPass<P,NextMove<P> >
    (new_hash, move, helper);
  this->eval = old_eval;
  if (move.isPass())
    this->pass_count.dec(P);

  return result;
}

template <class EvalT>
template <osl::Player P>
void osl::search::AlphaBeta2Tree<EvalT>::
examineMovesRoot(const MoveLogProbVector& moves, size_t i, Window window,
		 MoveLogProb& best_move, int& best_value)
{
  for (;i<moves.size(); ++i) {
    testStop();

#if (defined OSL_SMP) && (! defined OSL_SMP_NO_SPLIT_ROOT)
    if (shared && i > 8
	&& moves.size() > i+1) {
      int smp_idle;
      {
#  ifdef OSL_USE_RACE_DETECTOR
	boost::mutex::scoped_lock lk(shared->lock_smp);
#  endif
	smp_idle = shared->smp_idle;
      }
      if (smp_idle) {
	try {
	  examineMovesRootPar<P>(moves, i, window, best_move, best_value);
	  break;
	} catch (AlphaBeta2ParallelCommon::SplitFailed&) {
	}
      }
    }
#endif

    const MoveLogProb& m = moves[i];
#ifndef MINIMAL
    if (OslConfig::usiMode())
      UsiReporter::rootMove(m.getMove());
#endif
    const int result = alphaBetaSearch<P>(m, window, false);
    if (eval::betterThan(P, result, best_value)) {
      window.alpha(P) = result + EvalTraits<P>::delta;
      best_move = m;
      best_value = result;
      updateRootPV(P, std::cerr, result, m.getMove());
      if (eval::betterThan(P, result, window.beta(P))) {
	assert(! this->isWinValue(alt(P), result));
	break;
      }
    }
  }
}

/* ------------------------------------------------------------------------- */

template <class EvalT>
osl::search::AlphaBeta2<EvalT>::
AlphaBeta2(const NumEffectState& s, checkmate_t& c, 
	   SimpleHashTable *t, CountRecorder& r)
  : AlphaBeta2Tree<EvalT>(s, c, t, r)
{
  MoveGenerator::initOnce();
}

template <class EvalT>
osl::search::AlphaBeta2<EvalT>::
~AlphaBeta2()
{
}

template <class EvalT>
osl::Move osl::search::AlphaBeta2<EvalT>::
computeBestMoveIteratively(int limit, const int step, 
			   int initial_limit, int node_limit,
			   const TimeAssigned& assign,
			   MoveWithComment *additional_info)
{
  this->setStartTime(MilliSeconds::now());
  this->setTimeAssign(assign);
  if (this->table->isVerbose() > 2)
  {
    const time_t now = time(0);
    char ctime_buf[64];
    std::cerr << "AlphaBeta2 " << ctime_r(&now, ctime_buf);
  }
  if (this->table->isVerbose())
    std::cerr << " time assign/max " << this->timeAssigned().standard.toSeconds()
	      << "/" << this->timeAssigned().max.toSeconds() << "\n";
  initial_limit = std::min(initial_limit, limit);
  
  this->recorder.resetNodeCount();

  double last_iteration_consumed = 0;
  double total_consumed = 0;
  int limit_iterative = initial_limit;
  Move last_best_move = Move::INVALID();
  this->shared_root->last_pv.clear();

#ifdef OSL_SMP
#  ifdef SPLIT_STAT
  if (this->shared) {
    this->shared->parallel_splits = 0;
    this->shared->cancelled_splits.setValue(0);
    this->shared->parallel_abort.setValue(0);
  }
#  endif
#endif
  try
  {
    if (this->table->verboseLevel() > 1)
    {
      MoveVector moves;
      move_generator::LegalMoves::generate(this->state(), moves);
      for (size_t i=0; i<moves.size(); ++i) {
	HashKey key = this->currentHash().newHashWithMove(moves[i]);
	const SimpleHashRecord *record = this->table->find(key);
	if (! record || record->lowerLimit() < SearchTable::HistorySpecialDepth)
	  continue;
	std::cerr << "prebound value " << record::csa::show(moves[i])
		  << " " << record->lowerBound() << " " << record->upperBound() << "\n";
      }
    }

    MoveLogProb search_move;
    this->shared_root->last_root_value.push_back(alphaBetaSearchRoot(search_move, 0));
    this->shared_root->last_root_move = search_move.getMove();
    if (this->table->verboseLevel() > 1)
      std::cerr << "=> quiesce "
		<< record::csa::show(search_move.getMove()) << "\n";
    while (limit_iterative < limit)
    {
      if (this->table->verboseLevel() > 1)
	std::cerr << "=> iteration " << limit_iterative 
		  << " (" << last_iteration_consumed << ", " << total_consumed << ")"  << "\n";
      this->recorder.startSearch(limit_iterative);
      const int previous_node_count = this->nodeCount();
      try {
	this->shared_root->last_root_value.push_back(alphaBetaSearchRoot(search_move, limit_iterative));
	this->shared_root->last_root_move = search_move.getMove();
      } catch (...) {
	last_iteration_consumed = this->elapsed() - total_consumed;
	total_consumed += last_iteration_consumed;
	this->updateCheckmateCount();
	this->recorder.finishSearch(search_move.getMove(), total_consumed,
			      this->table->verboseLevel());
	throw;
      }

      last_best_move = search_move.getMove();
      last_iteration_consumed = this->elapsed() - total_consumed;
      total_consumed += last_iteration_consumed;

      this->updateCheckmateCount();
      if (this->table->verboseLevel() > 2) {
	std::cerr << "<= " 
		  << record::csa::show(search_move.getMove());
	std::cerr << std::setprecision(4) << "  mpn " << this->mpn.getAverage() 
		  << " cut " << this->mpn_cut.getAverage()
		  << " alpha " << this->alpha_update.getAverage()
		  << " last " << this->last_alpha_update.getAverage()
		  << " ext " << 100.0*this->ext.getAverage() << "%"
		  << " ext_limit " << this->ext_limit.getAverage();
#ifdef OSL_SMP
#  ifdef SPLIT_STAT
	if (this->shared) {
	  std::cerr << " split " << this->shared->parallel_splits << " cancel " << this->shared->cancelled_splits.value() 
		    << " abort " << this->shared->parallel_abort.value();
	}
#  endif
#endif
	std::cerr << "\n";
      }
      bool time_over = false;
      if (this->hasSchedule()) {
	const double elapsed = this->elapsed();
	const double current_time_left = this->timeAssigned().standard.toSeconds() - elapsed;
	double coef = this->nextIterationCoefficient();
	if (! this->isStableNow())
	  coef = std::min(0.5, coef);
	if (current_time_left 
	    < last_iteration_consumed * coef)
	  time_over = true;
	if (! time_over) {
	  SimpleHashRecord *record
	    = this->table->find(this->currentHash());
	  if (record) {
	    record->addNodeCount(this->nodeCount() - previous_node_count);
	  }
	}
      }
#ifdef OSL_SMP
      const size_t main_checkmate_count
	= this->shared ? this->shared->mainCheckmateCount() : this->checkmateSearcher().mainNodeCount();
#else
      const size_t main_checkmate_count = this->checkmateSearcher().mainNodeCount();
#endif
      const bool checkmate_over = main_checkmate_count > CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT;
      bool node_limit_over = (this->recorder.nodeCount() *4 > node_limit)
	|| checkmate_over;
      this->recorder.finishSearch(search_move.getMove(), 
			    total_consumed,
			    (time_over || node_limit_over) && this->table->verboseLevel());
      if (time_over || node_limit_over) {
	if (this->table->isVerbose()) {
	  std::cerr << "iteration stop at " << limit_iterative << " by "
		    << (time_over ? "time" : (checkmate_over ? "checkmate count" : "node count")) << "\n";
	  if (checkmate_over)
	    std::cerr << "checkmate cout " << main_checkmate_count << " > " << CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT << "\n";
	}
	goto finish;
      }
      this->testStop();
      limit_iterative += step;
    }
    if (this->table->verboseLevel() > 1)
      std::cerr << "=> final iteration " << limit_iterative 
		<< " (" << last_iteration_consumed << ", " << total_consumed << ")"  << "\n";
    while (true) {
      this->recorder.startSearch(limit);
      try {
	this->shared_root->last_root_value.push_back(alphaBetaSearchRoot(search_move, limit));
	this->shared_root->last_root_move = search_move.getMove();
      } catch (...) {
	last_iteration_consumed = this->elapsed() - total_consumed;
	total_consumed += last_iteration_consumed;
	this->updateCheckmateCount();
	this->recorder.finishSearch(search_move.getMove(), total_consumed,
			      this->table->verboseLevel());
	throw;
      }
      last_iteration_consumed = this->elapsed() - total_consumed;
      total_consumed += last_iteration_consumed;
      this->updateCheckmateCount();
      this->recorder.finishSearch(search_move.getMove(), total_consumed,
			    this->table->verboseLevel());
      last_best_move = search_move.getMove();

      if (last_best_move.isNormal())
	break;
      this->testStop();

      // ۤäƤλ
      if (limit >= 2000)
	break;

      limit += 200;
      if (this->table->isVerbose())
	std::cerr << "  extend limit to " << limit << " before resign\n";
    }
  }
  catch (std::exception& e)
  {
    if (! OslConfig::usiMode())
      std::cerr << "std exception " << e.what() << "\n";
  }
  catch (...)
  {
    std::cerr << "unknown exception\n";
#ifndef NDEBUG
    throw;
#endif
  }
finish:
  if (this->table->verboseLevel() > 1) {
    std::cerr << "<= " << record::csa::show(last_best_move);
    std::cerr << std::setprecision(4) << "  mpn " << this->mpn.getAverage()
	      << " cut " << this->mpn_cut.getAverage()
	      << " alpha " << this->alpha_update.getAverage()
	      << " last " << this->last_alpha_update.getAverage()
	      << " ext " << this->ext.getAverage()
	      << " ext_limit " << this->ext_limit.getAverage();
#ifdef OSL_SMP
#  ifdef SPLIT_STAT
    if (this->shared) {
      std::cerr << " split " << this->shared->parallel_splits << " cancel " << this->shared->cancelled_splits.value() 
		<< " abort " << this->shared->parallel_abort.value();
    }
#  endif
#endif
    std::cerr << "\n";
  }

  if (additional_info && this->shared_root->last_root_value.size() > 1) { // last_root_value[0] is for quiesce
    assert(last_best_move == this->shared_root->last_root_move);
    additional_info->move = last_best_move;
    additional_info->moves.clear();
    const double scale = 200.0/this->eval.captureValue(newPtypeO(WHITE,PAWN));
    additional_info->value = static_cast<int>(this->shared_root->last_root_value_update * scale);
    for (size_t i=1; i<this->shared_root->last_pv.size(); ++i) {
      additional_info->moves.push_back(this->shared_root->last_pv[i]);
    }
  }
  return last_best_move;
}

template <class EvalT>
template <osl::Player P>
int osl::search::AlphaBeta2<EvalT>::
alphaBetaSearchRoot(Window window, MoveLogProb& best_move, int limit)
{
#ifndef MINIMAL
  if (OslConfig::usiMode())
    UsiReporter::newDepth(limit/200);
#endif
  assert(P == this->state().getTurn());
  assert(window.alpha(P) % 2);
  assert(window.beta(P) % 2);
  setRoot(limit);
  assert(this->curDepth() == 0);
  this->node_type[this->curDepth()] = base_t::PvNode;
#ifdef NEW_DFPN
  this->checkmate_searcher->setRootPlayer(P);
#endif
#ifdef OSL_SMP
  if (this->shared)
    this->shared->threadStart();
#endif
  // ޤơ֥򸣤
  SimpleHashRecord *record_in_table
    = this->table->allocate(this->currentHash(), limit);
  SimpleHashRecord *record = record_in_table;
  boost::scoped_ptr<SimpleHashRecord> record_if_not_allocated;
  if (! record)
  {
    record_if_not_allocated.reset(new SimpleHashRecord());
    record = record_if_not_allocated.get();
  }
  assert(record);
  this->setRootRecord(record);
  assert(this->rootRecord() == record);
  assert(this->hasLastRecord() && this->lastRecord() == record);
  record->setInCheck(this->state().inCheck());

  if (limit == 0) {
    int result = this->template quiesce<P>(fullWindow(P));
    best_move = MoveLogProb(record->qrecord.bestMove(), 100);
    if (this->root_ignore_moves
	&& this->root_ignore_moves->isMember(best_move.getMove()))
      best_move = MoveLogProb();
    return result;
  }
  if (record_in_table) {
    int table_value = 0;
    const MoveLogProb m = record_in_table->bestMove().moveLogProb();
    if (! m.getMove().isNormal())
      record_in_table->resetValue();
    else if (record->hasGreaterLowerBound<P>(this->curLimit(), window.beta(P), 
					     table_value)) {
      if (! this->root_ignore_moves 
	  || ! this->root_ignore_moves->isMember(m.getMove())) {
	best_move = m;
	return table_value;
      }
    }
  }

  // gather all moves
  MoveLogProbVector moves;
  MoveGenerator& generator = this->makeGenerator();
  const MoveLogProb last_best_move = record->bestMove().moveLogProb();
  {
    MoveLogProbVector raw_moves;
    assert(this->curLimit() > 0);
    generator.init(this->curLimit()+200, record, this->eval, this->state(), true, last_best_move.getMove());
    if (last_best_move.getMove().isNormal())
      raw_moves.push_back(last_best_move);
    else if (record->qrecord.bestMove().isNormal())
      raw_moves.push_back(MoveLogProb(record->qrecord.bestMove(), 100));
    generator.generateAll<P>(*this, raw_moves);

    // clean up losing moves
    for (size_t i=0; i<raw_moves.size(); ++i) {
      const Move m = raw_moves[i].getMove();
      const HashKey key = this->currentHash().newHashWithMove(m);
      const SimpleHashRecord *record = this->table->find(key);
      if (record) {
	if (record->hasLowerBound(SearchTable::HistorySpecialDepth)
	    && this->isWinValue(alt(P), record->lowerBound()))
	  continue;
      }
      if (this->root_ignore_moves && this->root_ignore_moves->isMember(m))
	continue;
      if (! m.isDrop() && m.ptype() != KING
	  && move_classifier::KingOpenMove<P>::isMember(this->state(), m.ptype(), m.from(), m.to()))
	continue;

      raw_moves[i].setLogProbAtMost(limit);
      moves.push_back(raw_moves[i]);
    }
  }

  if (moves.size() == 1 
      || (moves.size() == 2 && moves[0].getMove() == moves[1].getMove()))
  {
    best_move = moves[0];
    return (window.alpha(P)+window.beta(P))/2;	// XXX
  }

#ifndef DONT_USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  int checkmate_node = 0;
  if (this->root_ignore_moves == 0) {
    int checkmate_max = 30000*std::max(limit - 300, 0)/100;
    if (limit >= 1000)
      checkmate_max = std::min(400000, 60000*(limit - 800)/100);
    checkmate_node = record->qrecord.checkmateNodesLeft(checkmate_max);
#ifdef CHECKMATE_COUNT
    std::cerr << "limit " << limit << " checkmate " << checkmate_node << "\n";
#endif
  }
  if (checkmate_node > 0)
  {
    const bool my_king_in_check
      = this->state().hasEffectBy(alt(P),this->state().getKingPosition(P));
    if (my_king_in_check)
    {
      // ꤫鲦꤬äƤ
      this->recorder.gotoCheckmateSearch(this->state(), checkmate_node/8);
      const bool lose = this->template isLosingState<P>(checkmate_node/8);
      this->recorder.backFromCheckmateSearch();
      this->updateCheckmateCount();
      if (lose)
      {
	best_move = MoveLogProb(Move::INVALID(),100);
	this->recordLoseByCheckmate(P, record);
	this->shared_root->last_pv.clear();
	this->shared_root->last_root_move = Move();
	this->shared_root->last_root_value_update = this->winByCheckmate(alt(P));
	return this->winByCheckmate(alt(P));
      }
    }
    // ͤޤʤСͤޤȤͤ 
    {
      Move checkmate_move;
#ifdef CHECKMATE_COUNT
      size_t count = checkmateSearcher().totalNodeCount();
#endif
      this->recorder.gotoCheckmateSearch(this->state(), checkmate_node);
      const bool win = this->template isWinningState<P>
	(checkmate_node, checkmate_move, record->qrecord.attack_oracle);
      this->recorder.backFromCheckmateSearch();
      this->updateCheckmateCount();
#ifdef CHECKMATE_COUNT
      root_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif
      if (win)
      {
	best_move = MoveLogProb(checkmate_move,100);
	this->recordWinByCheckmate(P, record, checkmate_move);
	this->shared_root->last_pv.clear();
	this->shared_root->last_root_move = checkmate_move;
	this->shared_root->last_root_value_update = this->winByCheckmate(P);
	this->pv[1].clear();
	this->updateRootPV(P, std::cerr, this->winByCheckmate(P), checkmate_move);
	return this->winByCheckmate(P);
      }
    }
    // ͤͤ
    if ((! my_king_in_check)
	&& (! (record->threatmate().isThreatmate(P))))
    {
      Move threatmate_move;
#ifdef CHECKMATE_COUNT
      size_t count = checkmateSearcher().totalNodeCount();
#endif
      this->recorder.gotoCheckmateSearch(this->state(), checkmate_node);
      const bool threatmate 
	= this->template isThreatmateState<P>
	(checkmate_node, threatmate_move, record->qrecord.threatmate_oracle);
#ifdef CHECKMATE_COUNT
      root_checkmate += checkmateSearcher().totalNodeCount() - count;
#endif
      this->recorder.backFromCheckmateSearch();
      this->updateCheckmateCount();
      if (threatmate)
      {
	if (record)
	  record->threatmate().setThreatmate(P, threatmate_move);
	if (this->table->verboseLevel() > 1)
	  std::cerr << "  root threatmate " << threatmate_move << "\n";
      }
    }

  }
#endif

  const int ValueNone = window.alpha(P) - EvalTraits<P>::delta;
  int best_value = ValueNone;
  try {
    // first move
    size_t i=0;
    if (limit >= 1000)
    {
      // try asipration window if we have sufficient limit
      const int root_alpha = this->rootAlpha(P, this->shared_root->last_root_value.size() ? this->shared_root->last_root_value.back() : 0);
      if (EvalTraits<P>::betterThan(root_alpha, window.alpha(P))) {
	const Window window_copy = window;
	window.alpha(P) = root_alpha;
	const int result = this->template alphaBetaSearch<P>(moves[0], window, true);
	if (EvalTraits<P>::betterThan(result, root_alpha))
	{
	  window.alpha(P) = result + EvalTraits<P>::delta;
	  best_move = moves[0];
	  best_value = result;
	  this->updateRootPV(P, std::cerr, result, moves[0].getMove());
	  ++i;
	} 
	else
	{
	  if (this->table->isVerbose())
	    this->showFailLow(result, moves[0].getMove());
	  this->setStable(false);
	  window = window_copy;
	}
      }
    }
    for (;i<moves.size() && best_value == ValueNone; ++i) {
      const MoveLogProb& m = moves[i];
#ifndef MINIMAL
      if (OslConfig::usiMode())
	UsiReporter::rootMove(m.getMove());
#endif
      const int result = this->template alphaBetaSearch<P>(m, window, true);
      if (eval::betterThan(P, result, best_value)) {
	window.alpha(P) = result + EvalTraits<P>::delta;
	best_move = m;
	best_value = result;
	this->updateRootPV(P, std::cerr, result, m.getMove());
	if (eval::betterThan(P, result, window.beta(P))) {
	  assert(! this->isWinValue(alt(P), result));
	}
      }
    }
    // other moves
    if (! eval::betterThan(P, window.alpha(P), window.beta(P))) {
      this->template examineMovesRoot<P>(moves, i, window, best_move, best_value);
    }
    if (best_move.getMove().isNormal()) {
      if (best_value != ValueNone) {
	assert(! this->shared_root->last_pv.empty());
	assert(best_move.getMove() == this->shared_root->last_pv[0]);
      }
    }
  } catch (misc::NoMoreTime&) {
    assert(best_value % 2 == 0);
    if (best_value != ValueNone)
      record->setLowerBound(P, this->curLimit(), SearchMove(best_move),
			    best_value);
    if (best_move.validMove()
	&& best_move.getMove() != last_best_move.getMove()) {
      if (this->table->verboseLevel() > 1) {
	std::cerr << "! use better move than the last best move\n";
	if (best_value != ValueNone) {
	  assert(! this->shared_root->last_pv.empty());
	  assert(best_move.getMove() == this->shared_root->last_pv[0]);
	}
      }
    }
    else {
#ifdef OSL_SMP
      if (this->shared)
	this->shared->waitAll();
#endif      
      throw;
    }
  }
  
  assert(best_value % 2 == 0);
  if (best_value != ValueNone)
    record->setLowerBound(P, this->curLimit(), SearchMove(best_move),
			  best_value);
#ifdef OSL_SMP
  if (this->shared)
    this->shared->waitAll();
#endif      
  return best_value;
}

template <class EvalT>
void osl::search::AlphaBeta2<EvalT>::setRoot(int limit)
{
  SearchState2::setRoot(limit);
  SimpleHashRecord *record = this->table->allocate(this->currentHash(), std::max(1000,limit));
  assert(record);
  this->setRootRecord(record);  
  this->move_type[this->curDepth()] = base_t::INITIAL;
}

template <class EvalT>
void osl::search::AlphaBeta2<EvalT>::makeMove(Move move)
{
  assert(this->state().isValidMove(move));
  SearchState2::makeMove(move);
  this->eval.update(this->state(), move);

  SimpleHashRecord *record 
    = this->table->allocate(this->currentHash(), this->curLimit());
  assert(record);
  this->move_type[this->curDepth()] = base_t::INITIAL;
  record->setInCheck(this->state().inCheck());
  this->setCurrentRecord(record);
}

template <class EvalT>
bool osl::search::AlphaBeta2<EvalT>::
isReasonableMove(Move /*move*/, int /*pawn_sacrifice*/)
{
  return true;
}

template <class EvalT>
void osl::search::AlphaBeta2<EvalT>::
showNodeDepth(std::ostream& os)
{
  int max_depth=0;
  for (int i=base_t::MaxDepth-1; i>=0; --i) {
    if (base_t::depth_node_count[i] || base_t::depth_node_count_quiesce[i]) {
      max_depth = i;
      break;
    }
  }
  int max_count=0;
  for (int i=0; i<=max_depth; i+=2) {
    max_count = std::max(max_count, 
			 base_t::depth_node_count[i]+base_t::depth_node_count_quiesce[i]);
  }

  int unit = std::max(max_count/79, 100);
  for (int i=0; i<=max_depth; i+=2) {
    os << std::setw(3) << i << " " 
       << std::string(base_t::depth_node_count[i]/unit, '*')
       << std::string(base_t::depth_node_count_quiesce[i]/unit, '+')
       << std::endl;
  }
#ifdef CHECKMATE_COUNT
  std::cerr << "checkmate root " << root_checkmate << " quiesce " << quiesce_checkmate
	    << "\nnormal before " << checkmate_before 
	    << " after " << checkmate_after << " threatmate " << count_threatmate
	    << "\n";
#endif
}

template <class EvalT>
void osl::search::AlphaBeta2<EvalT>::
clearNodeDepth()
{
  base_t::depth_node_count.fill(0);
  base_t::depth_node_count_quiesce.fill(0);  
}

namespace osl
{
  namespace search
  {
#ifndef MINIMAL
    template class AlphaBeta2<eval::ProgressEval>;
    template class AlphaBeta2Tree<eval::ProgressEval>;
    template
    void AlphaBeta2Tree<eval::ProgressEval>::examineMovesRoot<BLACK>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
    template
    void AlphaBeta2Tree<eval::ProgressEval>::examineMovesRoot<WHITE>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
#endif
    template class AlphaBeta2<eval::experimental::TestEval>;
    template class AlphaBeta2Tree<eval::experimental::TestEval>;
    template
    void AlphaBeta2Tree<eval::experimental::TestEval>::examineMovesRoot<BLACK>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
    template
    void AlphaBeta2Tree<eval::experimental::TestEval>::examineMovesRoot<WHITE>(const MoveLogProbVector&, size_t, Window, MoveLogProb&, int&);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
