#include "osl/search/searchMoveVector.h"
#include "osl/search/searchMove.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/eval/evalTraits.h"
#include "osl/misc/fixedCapacityVector.h"
#include "osl/misc/reorder.h"
#include <algorithm>
#include <iostream>

#ifndef MINIMAL
std::ostream& osl::search::operator<<(std::ostream& os,
				      SearchMoveVector const& mv)
{
  os<< "SearchMoveVector" << std::endl;
  for (SearchMoveVector::const_iterator p=mv.begin(); p!=mv.end(); ++p)
  {
    os << (*p)->moveLogProb() << " " << (*p)->record << std::endl;
  }
  return os << std::endl;
}
#endif
namespace osl
{
  namespace search
  {
    template <bool prefer_less>
    struct LogProbCompare
    {
      bool operator()(const SearchMove *l, const SearchMove *r) const
      {
	if (prefer_less)
	  return l->getLogProb() < r->getLogProb();
	else
	  return l->getLogProb() > r->getLogProb();
      }
    };

    template <Player P>
    struct SearchValueCompare
    {
      const int default_value;
      explicit SearchValueCompare(int d) : default_value(d) {}
      int value(const SearchMove& move) const
      {
	const SimpleHashRecord *record = move.record;
	if (! record)
	  return default_value;
	if (record->hasUpperBound(0))
	{
	  const int upper = record->upperBound();
	  if (record->hasLowerBound(0))
	    return EvalTraits<P>::max(upper, record->lowerBound());
	  return upper;
	}
	return (record->hasLowerBound(0) ? record->lowerBound() : default_value);
      }
      bool operator()(const SearchMove *l, const SearchMove *r) const
      {
	if (l->record == 0 && r->record == 0)
	{
	  if (l->getLogProb() != r->getLogProb())
	    return l->getLogProb() < r->getLogProb();
	}
	const int l_value = value(*l);
	const int r_value = value(*r);
	return EvalTraits<P>::betterThan(l_value, r_value);
      }
    };
    template <Player P>
    struct SearchValueIndexCompare
    {
      typedef FixedCapacityVector<std::pair<int,int>,Move::MaxUniqMoves> vector_t;
      vector_t *values;
      const int default_value;
      
      explicit SearchValueIndexCompare(vector_t *v, int d) 
	: values(v), default_value(d) {}
      int value(int index) const { return (*values)[index].first; }
      int logProb(int index) const { return (*values)[index].second; }
      bool operator()(int l, int r) const
      {
	const int l_value = value(l);
	const int r_value = value(r);
	if (l_value == default_value && r_value == default_value)
	{
	  if (logProb(l) != logProb(r))
	    return logProb(l) < logProb(r);
	}	
	return EvalTraits<P>::betterThan(l_value, r_value);
      }
    };

    template void SearchMoveVector::sortByValue<BLACK>(int);
    template void SearchMoveVector::sortByValue<WHITE>(int);
  }
}

void osl::search::
SearchMoveVector::sortByProbability()
{
  std::sort(begin(), end(), LogProbCompare<true>());
}
void osl::search::
SearchMoveVector::sortByProbabilityReverse()
{
  std::sort(begin(), end(), LogProbCompare<false>());
}

template <osl::Player P>
void osl::search::
SearchMoveVector::sortByValueUP(int default_value)
{
  // CAVEAT: this implementation is mt-unsafe
  std::sort(begin(), end(), SearchValueCompare<P>(default_value));
}

template <osl::Player P>
void osl::search::
SearchMoveVector::sortByValueMP(int default_value)
{
  if (empty())
    return;
  
  FixedCapacityVector<std::pair<int,int>,Move::MaxUniqMoves> values;
  FixedCapacityVector<int,Move::MaxUniqMoves> indices;
  SearchValueCompare<P> comp(default_value);
  for (const_iterator p=begin(); p!=end(); ++p) {
    values.push_back(std::make_pair(comp.value(**p), (*p)->getLogProb()));
    indices.push_back(indices.size());
  }
  std::sort(indices.begin(), indices.end(), 
	    SearchValueIndexCompare<P>(&values, default_value));

  misc::Reorder::reorder(begin(), end(), indices);
}

template <osl::Player P>
void osl::search::
SearchMoveVector::sortByValue(int default_value)
{
#ifndef OSL_SMP
  sortByValueUP<P>(default_value);
#else
  sortByValueMP<P>(default_value);
#endif
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
